#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <errno.h>

#include <iostream.h>
#include <string.h>

#include "sys_msg.hh"

#ifdef __sun__
#include "b.hh"
#define mtyp_t  long
extern "C" void perror(const char *s);
#endif

#ifdef __linux__
#define mtyp_t  long
#endif


SYS_Message_Queue::SYS_Message_Queue(int q,int f)
: mq(-1), flag(f)
{
  msqid_ds  mqd;

  if((flag & SYS_MSGQUEUE_LINK)&&(flag & SYS_MSGQUEUE_CREATE))
    {
      cerr << "bad incompatible options for msgqueue\n";
    }
  else if((q == SYS_MSGQUEUE_ANY)&&(flag & SYS_MSGQUEUE_LINK))
    {
      cerr << "bad request for msgqueue\n";
    }
  else
    {
      if(q == SYS_MSGQUEUE_ANY)
	q = IPC_PRIVATE;
      
      if(flag & SYS_MSGQUEUE_CREATE)
	{
	  //cout << "creating msgqueue " << q << "\n";
	  if((mq = msgget(q,IPC_CREAT|0777)) < 0)
	    {
	      cerr << "SYS QUEUE CREATE: no queue\n";
	      perror("SYS QUEUE CREATE");
	    }
	  else
	    {
	      //cout << "create msgqueue " << mq << "\n";
	      if(msgctl(mq,IPC_STAT,&mqd) < 0)
		{
		  cerr << "SYS QUEUE STAT: no queue\n";
		  perror("SYS QUEUE STAT");
		}
	      else
		{

		  //cout << "init get errno: " << errno << '\n';
		  //cout << "old msg_qbytes = " << mqd.msg_qbytes << '\n'; 
/*
		  mqd.msg_qbytes = 16384;
		  if(msgctl(mq,IPC_SET,&mqd) < 0)
		    {
		      cerr << "SYS QUEUE SET: no queue; errno = " << errno << '\n';
		      perror("SYS QUEUE SET");
		    }
		  //cout << "init set errno: " << errno << '\n';
*/
		}
	    }
	}
      else if(flag & SYS_MSGQUEUE_LINK)
	{
	  //cout << "linking with msgqueue " << q << "\n";
	  if((mq = msgget(q,0777)) < 0)
	    {
	      cerr << "SYS QUEUE LINK: bad queue" << q << '\n';
	      perror("SYS QUEUE LINK");
	    }
	}
    }
}


SYS_Message_Queue::~SYS_Message_Queue()
{
  if(mq != -1)
    {
      if(flag & SYS_MSGQUEUE_CREATE)
	{
	  //cout << "destroing queue: " << mq << "\n";
	  if(msgctl(mq,IPC_RMID,NULL)<0)
	    {
	      cerr << "SYS QUEUE ~(): errno = " << errno << '\n';
	      perror("SYS QUEUE ~()");
	    }
	}
      else if(flag & SYS_MSGQUEUE_LINK)
	{
	  ; //cout << "unlinking queue: " << mq << "\n";
	}
    }
}

#ifdef __osf__
int SYS_Message_Queue::QID()
{
  return mq;
}
#endif

#ifdef __sun__
int SYS_Message_Queue::QID()
{
  return mq;
}
#endif

#ifdef __linux__
int SYS_Message_Queue::QID()
{
  return mq;
}
#endif


int SYS_Message_Queue::Send(int type,void* buf,int len)
{
  if(this->mq == -1)
    return -1;

  struct msgbuf* pmb =(struct msgbuf*) new char[len+sizeof(mtyp_t)];
  pmb->mtype = type;
  bcopy((char*)buf,pmb->mtext,len);

  int ret = msgsnd(this->mq,pmb,len,0x00);
  if(ret < 0)
    cerr << "SYS_MSG Send errno: " << errno << '\n';
  delete pmb;

  return ret;
}


int SYS_Message_Queue::Receive(int type,void* buf,int buflen)
{
  if(this->mq == -1)
    return -1;

  struct msgbuf* pmb =(struct msgbuf*) new char[buflen+sizeof(mtyp_t)];
  
  //pmb->mtype = type;
  int ret = msgrcv(this->mq,pmb,buflen,type,IPC_NOWAIT);
  if(ret > 0)
    bcopy(pmb->mtext,(char*)buf,ret);
  else
    cerr << "SYS_MQ receive errno: " << errno << '\n';
  
  delete pmb;
  return ret;
}


int SYS_Message_Queue::ReceiveSync(int type,void* buf,int buflen,int flags)
{
  if(this->mq == -1)
    return -1;

  struct msgbuf* pmb =(struct msgbuf*) new char[buflen+sizeof(mtyp_t)];
  int ret;
  
  while(1)
    {
      ret = msgrcv(this->mq,pmb,buflen,type,0);
      if(ret > 0)
	{
	  bcopy(pmb->mtext,(char*)buf,ret);

	  delete pmb;
	  return ret;
	}
      else if(ret == 0)
	{
	  delete pmb;
	  return ret;
	}
      else
	if(flags & NO_INTERRUPT)
	  {
	    if(errno != EINTR)
	      {
		cerr << "SYS_MQ receive sync errno: " << errno << "; ret = " << ret << " ; abort\n";
		delete pmb;
		return -1;
	      } 
	  }     
	else
	  {
	    delete pmb;
	    return -1;
	  }
    } 
}
